# To install, run 
# python setup.py build_ext -i
# Ref: https://github.com/pytorch/pytorch/blob/11a40410e755b1fe74efe9eaa635e7ba5712846b/test/cpp_extensions/setup.py#L62

from setuptools import setup
from torch.utils.cpp_extension import BuildExtension, CUDAExtension
import os

# USE_NINJA = os.getenv('USE_NINJA') == '1'
os.environ["CC"] = "gcc-7"
os.environ["CXX"] = "gcc-7"

USE_NINJA = os.getenv('USE_NINJA') == '1'

setup(
    name='standard_rasterize_cuda',
    ext_modules=[
	CUDAExtension('standard_rasterize_cuda', [
        'standard_rasterize_cuda.cpp',
        'standard_rasterize_cuda_kernel.cu',
        ])
	],
    cmdclass={'build_ext': BuildExtension.with_options(use_ninja=USE_NINJA)}
)



import torch
import numpy as np
from torch import nn
import torch.nn.functional as F
import numpy as np
from scipy import interpolate
from PIL import Image
from torchvision import transforms
from tqdm import tqdm

transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize(mean=[0.5, 0.5, 0.5],std=[0.5,0.5,0.5]),
])

class InputPadder:
    """ Pads images such that dimensions are divisible by 8 """
    def __init__(self, dims, mode='sintel'):
        self.ht, self.wd = dims[-2:]
        pad_ht = (((self.ht // 8) + 1) * 8 - self.ht) % 8
        pad_wd = (((self.wd // 8) + 1) * 8 - self.wd) % 8
        if mode == 'sintel':
            self._pad = [pad_wd//2, pad_wd - pad_wd//2, pad_ht//2, pad_ht - pad_ht//2]
        else:
            self._pad = [pad_wd//2, pad_wd - pad_wd//2, 0, pad_ht]

    def pad(self, *inputs):
        return [F.pad(x, self._pad, mode='replicate') for x in inputs]

    def unpad(self,x):
        ht, wd = x.shape[-2:]
        c = [self._pad[2], ht-self._pad[3], self._pad[0], wd-self._pad[1]]
        return x[..., c[0]:c[1], c[2]:c[3]]

def forward_interpolate(flow):
    flow = flow.detach().cpu().numpy()
    dx, dy = flow[0], flow[1]

    ht, wd = dx.shape
    x0, y0 = np.meshgrid(np.arange(wd), np.arange(ht))

    x1 = x0 + dx
    y1 = y0 + dy
    
    x1 = x1.reshape(-1)
    y1 = y1.reshape(-1)
    dx = dx.reshape(-1)
    dy = dy.reshape(-1)

    valid = (x1 > 0) & (x1 < wd) & (y1 > 0) & (y1 < ht)
    x1 = x1[valid]
    y1 = y1[valid]
    dx = dx[valid]
    dy = dy[valid]

    flow_x = interpolate.griddata(
        (x1, y1), dx, (x0, y0), method='nearest', fill_value=0)

    flow_y = interpolate.griddata(
        (x1, y1), dy, (x0, y0), method='nearest', fill_value=0)

    flow = np.stack([flow_x, flow_y], axis=0)
    return torch.from_numpy(flow).float()


def bilinear_sampler(img, coords, mode='bilinear', mask=False):
    """ Wrapper for grid_sample, uses pixel coordinates """
    H, W = img.shape[-2:]
    xgrid, ygrid = coords.split([1,1], dim=-1)
    xgrid = 2*xgrid/(W-1) - 1
    ygrid = 2*ygrid/(H-1) - 1

    grid = torch.cat([xgrid, ygrid], dim=-1)
    img = F.grid_sample(img, grid, align_corners=True)

    if mask:
        mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
        return img, mask.float()

    return img


def coords_grid(batch, ht, wd, device):
    coords = torch.meshgrid(torch.arange(ht, device=device), torch.arange(wd, device=device))
    coords = torch.stack(coords[::-1], dim=0).float()
    return coords[None].repeat(batch, 1, 1, 1)


def upflow8(flow, mode='bilinear'):
    new_size = (8 * flow.shape[2], 8 * flow.shape[3])
    return  8 * F.interpolate(flow, size=new_size, mode=mode, align_corners=True)



def get_batch_frames(frame_path_chunk, side_window, size):
    frame_path_chunk = frame_path_chunk[1:side_window+1][::-1] + frame_path_chunk + frame_path_chunk[-side_window-1:-1][::-1]
    
    batch_frames = []
    for frame_path in tqdm(frame_path_chunk):
        batch_frames += [transform(Image.open(frame_path).resize((size,size))).unsqueeze(0).cpu()]
    
    # enlarge frames for more accurate parsing maps and optical flows     
    # batch_frames = torch.cat(batch_frames, dim=0)
    batch_frames = torch.cat(batch_frames, dim=0)
    
    return batch_frames

def get_batch_labels(batch_frames, parsingpredictor, side_window):
    with torch.no_grad():
        batch_labels = parsingpredictor.parsing_net(2*batch_frames.to('cuda')).detach().cpu()
    batch_labels
    
    return batch_labels
    
    

def arrange_mask(parsing, output_size=512):
    Lbrow_mask = torch.where(parsing==2, 1, 0) + torch.where(parsing==3, 1, 0)
    Lbrow_mask[:, :, :output_size//2] = 0
    parsing = Lbrow_mask * 2 + (1 - Lbrow_mask) * parsing

    Rbrow_mask = torch.where(parsing==2, 1, 0) + torch.where(parsing==3, 1, 0)
    Rbrow_mask[:, :, output_size//2:] = 0
    parsing = Rbrow_mask * 3 + (1 - Rbrow_mask) * parsing

    Leye_mask = torch.where(parsing==4, 1, 0) + torch.where(parsing==5, 1, 0)
    Leye_mask[:, :, :output_size//2] = 0
    parsing = Leye_mask * 4 + (1 - Leye_mask) * parsing
    
    Reye_mask = torch.where(parsing==4, 1, 0) + torch.where(parsing==5, 1, 0)
    Reye_mask[:, :, output_size//2:] = 0
    parsing = Reye_mask * 5 + (1 - Reye_mask) * parsing
    
    Leye_mask = torch.where(parsing==7, 1, 0) + torch.where(parsing==8, 1, 0)
    Leye_mask[:, :, :output_size//2] = 0
    parsing = Leye_mask * 7 + (1 - Leye_mask) * parsing
    
    # Reye
    Reye_mask = torch.where(parsing==7, 1, 0) + torch.where(parsing==8, 1, 0)
    Reye_mask[:, :, output_size//2:] = 0
    parsing = Reye_mask * 8 + (1 - Reye_mask) * parsing
    return parsing
    
def get_innerface_mask(numpy_label):
    canvas = np.zeros_like(numpy_label)
    canvas += np.where(numpy_label==1, 1, 0)
    canvas += np.where(numpy_label==2, 1, 0)
    canvas += np.where(numpy_label==3, 1, 0)
    canvas += np.where(numpy_label==4, 1, 0)
    canvas += np.where(numpy_label==5, 1, 0)
    canvas += np.where(numpy_label==6, 1, 0)
    canvas += np.where(numpy_label==10, 1, 0)
    canvas += np.where(numpy_label==11, 1, 0)
    canvas += np.where(numpy_label==12, 1, 0)
    canvas += np.where(numpy_label==13, 1, 0)
    return canvas

def get_list_chunk(lists, n):
    m = len(lists)//n + 1 if len(lists)%n != 0 else len(lists)//n
    return [lists[i:i+m] for i in range(0,len(lists),m)]

# from RAFT
def warp(x, flo):
    """
    warp an image/tensor (im2) back to im1, according to the optical flow
    x: [B, C, H, W] (im2)
    flo: [B, 2, H, W] flow
    """
    B, C, H, W = x.size()
    # mesh grid 
    xx = torch.arange(0, W).view(1,-1).repeat(H,1)
    yy = torch.arange(0, H).view(-1,1).repeat(1,W)
    xx = xx.view(1,1,H,W).repeat(B,1,1,1)
    yy = yy.view(1,1,H,W).repeat(B,1,1,1)
    grid = torch.cat((xx,yy),1).float()


    #x = x.cuda()
    grid = grid.cuda()
    vgrid = grid + flo # B,2,H,W

    # scale grid to [-1,1] 
    ##2019 code
    vgrid[:,0,:,:] = 2.0*vgrid[:,0,:,:].clone()/max(W-1,1)-1.0 
    vgrid[:,1,:,:] = 2.0*vgrid[:,1,:,:].clone()/max(H-1,1)-1.0

    vgrid = vgrid.permute(0,2,3,1)
    output = nn.functional.grid_sample(x, vgrid,align_corners=True)
    mask = torch.autograd.Variable(torch.ones(x.size())).cuda()
    mask = nn.functional.grid_sample(mask, vgrid,align_corners=True)

    ##2019 author
    mask[mask<0.9999] = 0
    mask[mask>0] = 1

     ##2019 code
     # mask = torch.floor(torch.clamp(mask, 0 ,1))

    return output*mask, mask


import torch
from torch import nn

from MyModel.utils.spectral_norm import SpectralNorm
import torch.nn.functional as F

class Self_Attention(nn.Module):
    """ Self attention Layer"""

    def __init__(self, in_dim):
        super(Self_Attention, self).__init__()
        self.chanel_in = in_dim

        self.query_conv = SpectralNorm(nn.Conv2d(in_channels=in_dim, out_channels=in_dim // 1, kernel_size=1))
        self.key_conv = SpectralNorm(nn.Conv2d(in_channels=in_dim, out_channels=in_dim // 1, kernel_size=1))
        self.value_conv = SpectralNorm(nn.Conv2d(in_channels=in_dim, out_channels=in_dim, kernel_size=1))
        self.gamma = nn.Parameter(torch.zeros(1))

        self.softmax = nn.Softmax(dim=-1)  #

    def forward(self, x):
        """
        inputs :
            x : input feature maps( B X C X W X H)
        returns :
            out : self attention value + input feature
            attention: B X N X N (N is Width*Height)
        """
        m_batchsize, C, width, height = x.size()
        proj_query = self.query_conv(x).view(m_batchsize, -1, width * height).permute(0, 2, 1)  # B X CX(N)
        proj_key = self.key_conv(x).view(m_batchsize, -1, width * height)  # B X C x (*W*H)
        energy = torch.bmm(proj_query, proj_key)  # transpose check
        attention = self.softmax(energy)  # BX (N) X (N)
        proj_value = self.value_conv(x).view(m_batchsize, -1, width * height)  # B X C X N

        out = torch.bmm(proj_value, attention.permute(0, 2, 1))
        out = out.view(m_batchsize, C, width, height)

        out = self.gamma * out + x
        return out


class Discriminator_x64(nn.Module):
    """
    Discriminative Network
    """

    def __init__(self, in_size=6, ndf=64):
        super(Discriminator_x64, self).__init__()
        self.in_size = in_size
        self.ndf = ndf

        self.layer1 = nn.Sequential(
            SpectralNorm(nn.Conv2d(self.in_size, self.ndf, 4, 2, 1)), nn.LeakyReLU(0.2, inplace=True)
        )
        self.layer2 = nn.Sequential(
            SpectralNorm(nn.Conv2d(self.ndf, self.ndf, 4, 2, 1)),
            nn.InstanceNorm2d(self.ndf),
            nn.LeakyReLU(0.2, inplace=True),
        )
        self.attention = Self_Attention(self.ndf)
        self.layer3 = nn.Sequential(
            SpectralNorm(nn.Conv2d(self.ndf, self.ndf * 2, 4, 2, 1)),
            nn.InstanceNorm2d(self.ndf * 2),
            nn.LeakyReLU(0.2, inplace=True),
        )
        self.layer4 = nn.Sequential(
            SpectralNorm(nn.Conv2d(self.ndf * 2, self.ndf * 4, 4, 2, 1)),
            nn.InstanceNorm2d(self.ndf * 4),
            nn.LeakyReLU(0.2, inplace=True),
        )
        self.layer5 = nn.Sequential(
            SpectralNorm(nn.Conv2d(self.ndf * 4, self.ndf * 8, 4, 2, 1)),
            nn.InstanceNorm2d(self.ndf * 8),
            nn.LeakyReLU(0.2, inplace=True),
        )
        self.layer6 = nn.Sequential(
            SpectralNorm(nn.Conv2d(self.ndf * 8, self.ndf * 16, 4, 2, 1)),
            nn.InstanceNorm2d(self.ndf * 16),
            nn.LeakyReLU(0.2, inplace=True),
        )

        self.last = SpectralNorm(nn.Conv2d(self.ndf * 16, 1, [3, 6], 1, 0))

    def forward(self, input):
        feature1 = self.layer1(input)
        feature2 = self.layer2(feature1)
        feature_attention = self.attention(feature2)
        feature3 = self.layer3(feature_attention)
        feature4 = self.layer4(feature3)
        feature5 = self.layer5(feature4)
        feature6 = self.layer6(feature5)
        output = self.last(feature6)
        output = F.avg_pool2d(output, output.size()[2:]).view(output.size()[0], -1)

        return output, feature4



import cv2
import os
import numpy as np
import torch
import torch.nn.functional as F
import glob
import numpy as np
from PIL import Image
import PIL
import torch
from torchvision.transforms.functional import to_tensor
import shutil

def modulate_innerface(target_img, source_img, target_labels, source_labels):
    target_img = target_img.squeeze(0)
    source_img = source_img.squeeze(0)
    target_labels = target_labels[:,1:].sum(1)
    source_labels = source_labels[:,1:].sum(1)
    canvas = torch.zeros_like(target_img)
    for target_label, source_label in zip(target_labels, source_labels):
        target_mean, source_mean = 0, 0
        target_std, source_std = 1, 1
        if target_label.sum() and source_label.sum():
            target_pixels = torch.masked_select(target_img, target_label.bool())
            target_mean = target_pixels.mean()
            target_std = (target_pixels - target_mean).std() # , target_pixels.std()

            source_pixels = torch.masked_select(source_img, source_label.bool())
            source_mean = source_pixels.mean()
            source_std = (source_pixels - source_mean).std() # , target_pixels.std()
        
        edit_target_img = ((target_img - target_mean)/target_std) * source_std + source_mean
        
        canvas += edit_target_img * target_label
    
    return canvas.unsqueeze(0)

def modulate(target_img, source_img, target_labels, source_labels, skin_mean, skin_std, others_mean, others_std):
    target_img = target_img.squeeze(0)
    source_img = source_img.squeeze(0)
    target_labels = target_labels.squeeze(0)
    source_labels = source_labels.squeeze(0)
    
    # new skin(skin + nose + brow + eye + ear)
    target_labels[1, :, :] = (target_labels[1,:,:] + target_labels[2,:,:] + target_labels[3,:,:] + target_labels[4,:,:] + target_labels[5,:,:]\
         + target_labels[6,:,:] + target_labels[7,:,:] + target_labels[8,:,:] + target_labels[9,:,:] + target_labels[10,:,:]).clip(0,1)
    target_labels[2:11, :, :] *= 0
    
    source_labels[1, :, :] = (source_labels[1,:,:] + source_labels[2,:,:] + source_labels[3,:,:] + source_labels[4,:,:] + source_labels[5,:,:]\
        + source_labels[6,:,:] + source_labels[7,:,:] + source_labels[8,:,:] + source_labels[9,:,:] + source_labels[10,:,:]).clip(0,1)
    source_labels[2:11, :, :] *= 0
    
    # new lips(u lip + d lip)
    target_labels[12, :, :] = (target_labels[12,:,:] + target_labels[13,:,:] + target_labels[11,:,:]).clip(0,1)
    target_labels[13, :, :] *= 0
    target_labels[11, :, :] *= 0
    
    source_labels[12, :, :] = (source_labels[12,:,:] + source_labels[13,:,:] + source_labels[11,:,:]).clip(0,1)
    source_labels[13, :, :] *= 0
    source_labels[11, :, :] *= 0
    
    # mouth
    canvas = torch.zeros_like(target_img)
    for idx, (target_label, source_label) in enumerate(zip(target_labels[1:], source_labels[1:])):
        target_mean, source_mean = 0, 0
        target_std, source_std = 1, 1
        if target_label.sum() and source_label.sum():
            target_pixels = torch.masked_select(target_img, target_label.bool())
            target_mean = target_pixels.mean()
            target_std = (target_pixels - target_mean).std() # , target_pixels.std()

            source_pixels = torch.masked_select(source_img, source_label.bool())
            source_mean = source_pixels.mean()
            source_std = (source_pixels - source_mean).std() # , target_pixels.std()
        
            # if target_labels.shape[0] - 1 == idx:
            #     import pdb;pdb.set_trace()
            
            if idx == 0:
                edit_target_img = ((target_img - target_mean)/target_std)*source_std*skin_std + source_mean + skin_mean
                # edit_target_img = ((target_img - target_mean)/target_std)*source_std*.7 + source_mean + 0.2
            else:
                edit_target_img = ((target_img - target_mean)/target_std)*source_std*others_std + source_mean + others_mean
                # edit_target_img = ((target_img - target_mean)/target_std)*source_std*.7 + source_mean
                
            # if idx in [11, 12, 13]:
            #     target_label = target_label.cpu().numpy()[:,:,None]
            #     target_label = cv2.blur(target_label, (45, 45))
            #     target_label = torch.tensor(target_label, device='cuda').squeeze()
            blur_target_label = cv2.blur(np.array(target_label)[:,:,None], (13,13))
            canvas += edit_target_img * blur_target_label
    return canvas.unsqueeze(0)

# def modulate(target_img, source_img, target_labels, source_labels):
#     target_img = target_img.squeeze(0)
#     source_img = source_img.squeeze(0)
#     target_labels = target_labels.squeeze(0)
#     source_labels = source_labels.squeeze(0)
    
#     # nose + skin
#     target_labels[1, :, :] = target_labels[1,:,:] + target_labels[10,:,:]
#     target_labels[10, :, :] *= 0
    
#     source_labels[1, :, :] = source_labels[1,:,:] + source_labels[10,:,:]
#     source_labels[10, :, :] *= 0
    
#     canvas = torch.zeros_like(target_img)
#     for idx, (target_label, source_label) in enumerate(zip(target_labels[1:], source_labels[1:])):
#         target_mean, source_mean = 0, 0
#         target_std, source_std = 1, 1
#         if target_label.sum() and source_label.sum():
#             target_pixels = torch.masked_select(target_img, target_label.bool())
#             target_mean = target_pixels.mean()
#             target_std = (target_pixels - target_mean).std() # , target_pixels.std()

#             source_pixels = torch.masked_select(source_img, source_label.bool())
#             source_mean = source_pixels.mean()
#             source_std = (source_pixels - source_mean).std() # , target_pixels.std()
        
#         # if target_labels.shape[0] - 1 == idx:
#         #     import pdb;pdb.set_trace()
#         edit_target_img = ((target_img - target_mean)/target_std)*source_std + source_mean
#         # if idx in [11, 12, 13]:
#         #     target_label = target_label.cpu().numpy()[:,:,None]
#         #     target_label = cv2.blur(target_label, (45, 45))
#         #     target_label = torch.tensor(target_label, device='cuda').squeeze()
#         canvas += edit_target_img * target_label
    
#     return canvas.unsqueeze(0)

def to_one_hot(mask): # 0 ~ 8 h w
    mask_ = torch.tensor(mask, dtype=torch.int64)
    c = 19
    # c = np.array(list(face_parsing_converter.values())).max() + 1
    h,w = mask_.size()

    mask_ = torch.reshape(mask_,(1,1,h,w))
    one_hot_mask = torch.zeros(1, c, h, w)
    one_hot_mask_ = one_hot_mask.scatter_(1, mask_, 1.0)
    one_hot_mask_ = F.interpolate(one_hot_mask_, (h,w), mode='nearest')
    return one_hot_mask_

kernel = np.ones((3, 3), np.uint8)

def get_name_from_path(path):
    base_name = os.path.basename(path)
    frame_name, ext = os.path.splitext(base_name)
    return frame_name


def get_path_list(dir, sorting=True):
    if sorting:
        path_list = sorted(glob.glob(os.path.join(dir,'*.*')))
    else:
        path_list = glob.glob(os.path.join(dir,'*.*'))
        
    return path_list


def expand_and_transform_affine(points, affine_mat):
    constant_term = np.ones((106,1))
    expanded_points = np.concatenate((points, constant_term), axis=1)
    result = np.matmul(expanded_points, np.transpose(affine_mat))
    return result


def parse_list_to_batch(target_list, batch_size):
    batch_list = list(target_list[i:i+batch_size] for i in range(0, len(target_list), batch_size))
    return batch_list


def tensor_to_nparr(tensor, from_0=True):
    if from_0:
        numpy_255 = tensor.clone().detach().cpu().numpy().transpose([1,2,0]).clip(0,1)*255 
    else:
        numpy_255 = tensor.clone().detach().cpu().numpy().transpose([1,2,0]).clip(-1,1)*127.5+127.5
    
    return numpy_255
 
 
def label_to_mask_pair(np_label, size, kernel=kernel):
        np_label = np.where(np_label==8,1,np_label)
        np_label = np.where(np_label==7,1,np_label)
        
        # L_brow + R_brow
        np_label = np.where(np_label==3,2,np_label)
        # L_eye + R_eye
        np_label = np.where(np_label==5,4,np_label)
        # u_lip + d_lip + mouth(11)
        np_label = np.where(np_label==13,12,np_label)
        # np_label = np.where(np_label==11,12,np_label)
        
        
        innerface_mask = np.clip(np.where(np_label <= 13, 1, 0) - np.where(np_label == 0, 1, 0), 0, 1)
        dilation_mask = cv2.dilate(innerface_mask.astype(np.float32), kernel, iterations=25)
        boundary_mask = torch.tensor(dilation_mask - innerface_mask).view(1, 1, size, size)
        
        new_innerface_label = np_label * innerface_mask
        innerface_mask = torch.tensor(innerface_mask).view(1, 1, size, size)

        innerface_mask = torch.tensor(innerface_mask).view(1, 1, size, size)
        return innerface_mask, boundary_mask


def get_to_do(working_dir, standard_dir):
    working_len = len(get_path_list(working_dir, False)) 
    standard_len = len(get_path_list(standard_dir, False))
    
    if working_len == standard_len:
        return False
    
    else:
        return True
  
  
def spawn_forder(dir):
    num = len(os.listdir(dir))
    forder_name = str(num).zfill(4)
    new_forder_name = os.path.join(dir, forder_name)
    try:
        os.makedirs(new_forder_name)
        return new_forder_name
    
    except:
        raise Exception("폴더트리 꼬인듯?")
    
      
def find_lagacy(keyword, work_dir, root_work_dir):
    found_dir = glob.glob(os.path.join(root_work_dir, "*", keyword))
    if len(found_dir):
        for lagacy_dir in found_dir:
            if len(get_path_list(lagacy_dir)):
                return True, lagacy_dir

    return False, None
    
    
class Image_Probe():
    def __init__(self, object):
        self.object = object
        self.detect_type()
    
    def detect_type(self):
        if isinstance(self.object, PIL.Image.Image):
            self.data_type = 'pil'
            
        elif isinstance(self.object, np.ndarray):
            self.data_type = 'arr'
            
        elif isinstance(self.object, torch.Tensor):
            self.data_type = 'tensor'
            
        else:
            Exception(f"{self.save_name} 으로 저장하려는 객체는 이미지로 변환할 수 없습니다.")
        
    def visualize(self, path):
        if self.data_type == 'pil':
            self.object.save(path)
        
        if self.data_type == 'arr':
            Image.fromarray(self.object.astype(np.uint8)).save(path)
        
        if self.data_type == 'tensor':
            # print(type(list(self.object.size())))
            if len(list(self.object.size())) == 4:
                ch3_tensor = self.object[0]
            else: 
                ch3_tensor = self.object
        
            arr = ch3_tensor.cpu().clone().detach().numpy().transpose([1,2,0])*127.5+127.5
            cv2.imwrite(path, arr.astype(np.uint8))
            # Image.fromarray(arr.astype(np.uint8)).save(path)
            
            

"""
Label
"""            
def get_one_hot(mask): # 0 ~ 8 h w
    mask_ = torch.tensor(mask, dtype=torch.int64)
    c = 19
    if len(mask.shape) == 3:
        mask_ = mask_.unsqueeze(0)
    _,_,h,w = mask_.size()

    mask_ = torch.reshape(mask_,(1,1,h,w))
    one_hot_mask = torch.zeros((1, c, h, w), device=mask.device)
    one_hot_mask_ = one_hot_mask.scatter_(1, mask_, 1.0)
    return one_hot_mask_


def arrange_mask(parsing, output_size=512):
    Lbrow_mask = torch.where(parsing==2, 1, 0) + torch.where(parsing==3, 1, 0)
    Lbrow_mask[:, :, :output_size//2] = 0
    parsing = Lbrow_mask * 2 + (1 - Lbrow_mask) * parsing

    Rbrow_mask = torch.where(parsing==2, 1, 0) + torch.where(parsing==3, 1, 0)
    Rbrow_mask[:, :, output_size//2:] = 0
    parsing = Rbrow_mask * 3 + (1 - Rbrow_mask) * parsing

    Leye_mask = torch.where(parsing==4, 1, 0) + torch.where(parsing==5, 1, 0)
    Leye_mask[:, :, :output_size//2] = 0
    parsing = Leye_mask * 4 + (1 - Leye_mask) * parsing
    
    Reye_mask = torch.where(parsing==4, 1, 0) + torch.where(parsing==5, 1, 0)
    Reye_mask[:, :, output_size//2:] = 0
    parsing = Reye_mask * 5 + (1 - Reye_mask) * parsing
    
    Leye_mask = torch.where(parsing==7, 1, 0) + torch.where(parsing==8, 1, 0)
    Leye_mask[:, :, :output_size//2] = 0
    parsing = Leye_mask * 7 + (1 - Leye_mask) * parsing
    
    # Reye
    Reye_mask = torch.where(parsing==7, 1, 0) + torch.where(parsing==8, 1, 0)
    Reye_mask[:, :, output_size//2:] = 0
    parsing = Reye_mask * 8 + (1 - Reye_mask) * parsing
    return parsing

def get_convexhull_mask(pil_label, lmk):
    cv2_face = np.array(pil_label)
    canvas = np.zeros_like(cv2_face).astype(np.uint8)
    points = np.array([lmk[1], lmk[9], lmk[10], lmk[11], lmk[12], lmk[13], lmk[14], lmk[15], lmk[16], lmk[2], lmk[3], lmk[4], lmk[5], lmk[6], lmk[7], lmk[8], lmk[0], lmk[24], lmk[23], lmk[22], lmk[21], lmk[20], lmk[19], lmk[18], lmk[32], lmk[31], lmk[30], lmk[29], lmk[28], lmk[27], lmk[26], lmk[25], lmk[17], lmk[105], lmk[104], lmk[49], lmk[48]], np.int32)
    skin_mask = cv2.fillConvexPoly(canvas, points=points, color=(1,1,1))
    dilation_skin_mask = cv2.dilate(skin_mask, kernel, iterations=15)
    return dilation_skin_mask

def get_new_label(label, convexhull_mask):
    new_label = convexhull_mask * np.where(label==1,1,0) # skin and region
    new_label = new_label * (1 - np.where(label==2,1,0)) + np.where(label==2,1,0) * 2 # Lbrow
    new_label = new_label * (1 - np.where(label==3,1,0)) + np.where(label==3,1,0) * 3 # Rbrow
    new_label = new_label * (1 - np.where(label==4,1,0)) + np.where(label==4,1,0) * 4 # Leye
    new_label = new_label * (1 - np.where(label==5,1,0)) + np.where(label==5,1,0) * 5 # Reye
    new_label = new_label * (1 - np.where(label==6,1,0)) + np.where(label==6,1,0) * 6 # glasses
    new_label = new_label * (1 - np.where(label==10,1,0)) + np.where(label==10,1,0) * 10 # nose
    new_label = new_label * (1 - np.where(label==11,1,0)) + np.where(label==11,1,0) * 11 # mouth
    new_label = new_label * (1 - np.where(label==12,1,0)) + np.where(label==12,1,0) * 12 # ulips
    new_label = new_label * (1 - np.where(label==13,1,0)) + np.where(label==13,1,0) * 13 # dlips
    new_label = new_label * (1 - np.where(label==17,1,0)) # sub hair
    return new_label

    
"""
Face part
"""

facepart_lmk_indexes = {
    "L_eye" : [33, 35, 36, 37, 39, 40, 41, 42],
    "R_eye" : [87, 89, 90, 91, 93, 94, 95, 96],
    "mouth" : [52, 53, 55, 56, 58, 59, 61, 63, 64, 67, 68, 71]
}

def get_center_coord(lmks, part='L_eye'):
    assert part in facepart_lmk_indexes.keys()
    face_part_lmks = []
    for index in facepart_lmk_indexes[part]:
        face_part_lmks.append(np.array(lmks[index]))
    face_part_lmks = np.array(face_part_lmks)
    x_lmks, y_lmks = face_part_lmks[:,0], face_part_lmks[:,1]
    x_min, x_max = int(x_lmks.min()), int(x_lmks.max())
    y_min, y_max = int(y_lmks.min()), int(y_lmks.max())
    x_c, y_c = int(x_lmks.mean()), int(y_lmks.mean()) # sometimes (0,0)
    max_val = max(int(x_max - x_min), int(y_max - y_min))
    W, H = max_val, max_val
    
    return x_c, y_c, W, H


def get_part_img(aligned_face_512, aligned_lmk106, facepart, margin=16):
    x_c, y_c, W, H = get_center_coord(aligned_lmk106, facepart)

    crop_img = aligned_face_512[max(y_c-H//2-margin//2,0):min(y_c+H//2+margin//2,512), max(x_c-W//2-margin//2,0):min(x_c+W//2+margin//2, 512)]
    crop_img = np.array(crop_img, dtype=np.uint8)
    crop_img = cv2.resize(crop_img, (256,256))

    return crop_img

def get_grad_mask(size=256):
    x_axis = np.linspace(-1, 1, size)[:, None]
    y_axis = np.linspace(-1, 1, size)[None, :]

    arr1 = np.sqrt(x_axis ** 4 + y_axis ** 4)

    x_axis = np.linspace(-1, 1, size)[:, None]
    y_axis = np.linspace(-1, 1, size)[None, :]

    arr2 = np.sqrt(x_axis ** 2 + y_axis ** 2)

    grad_mask = np.clip(1-(arr1/2+arr2/2), 0, 1)
    return grad_mask



import cv2

from innerverz_package import Data_Process

DP = Data_Process()

image_path = 'sample.png'


"""
        face alignment usage
"""
from innerverz_package import FaceAligner
FA = FaceAligner(align_style='ffhq')

cv2_image = cv2.imread(image_path)
aligned_face, tfm, tfm_inv, lms_5, lms_106, face_bool = FA.get_face(cv2_image)

face_bool, lms_106, lms_5 = FA.detect_lmk(cv2_image)

aligned_face, tfm, tfm_inv = FA.align_face_from_5(cv2_image, lms_5)

"""
        ID extractor usage
"""
from innerverz_package import IdExtractor
IE = IdExtractor()

ts_aligned_face = DP.image_pp(aligned_face, 512, normalize=True)
id_vector = IE(ts_aligned_face)
similarity_score = IE.compare_similarity(id_vector, id_vector)



import cv2
import torch.nn.functional as F

from innerverz_package import Data_Process
from torchvision.transforms.functional import to_tensor, rgb_to_grayscale

image_path = 'sample.png'
DP = Data_Process()

"""
        face alignment usage
"""
from innerverz_package import FaceAligner
FA = FaceAligner(align_style='ffhq')

cv2_image = cv2.imread(image_path)
aligned_face, tfm, tfm_inv, lms_5, lms_106, face_bool = FA.get_face(cv2_image)

face_bool, lms_106, lms_5 = FA.detect_lmk(cv2_image)

aligned_face, tfm, tfm_inv = FA.align_face_from_5(cv2_image, lms_5)
aligned_lmk = FA.affine_transform(lms_106, tfm)

"""
        DECA usage
"""
from innerverz_package import DECA
deca = DECA()
image_dict = deca.data_preprocess(aligned_face, aligned_lmk)
code_dict, vis_dict = deca(image_dict)

inputs = DP.vis_pp(vis_dict['inputs'], normalize=False)
landmarks2d = DP.vis_pp(vis_dict['landmarks2d'], normalize=False)
landmarks3d = DP.vis_pp(vis_dict['landmarks3d'], normalize=False)
shape_images = DP.vis_pp(vis_dict['shape_images'], normalize=False)
shape_detail_images = DP.vis_pp(vis_dict['shape_detail_images'], normalize=False)


import os
import sys
import torch
import torch.nn as nn
import torch.nn.functional as F

from .nets import MyGenerator
from ..utils import check_ckpt_exist, get_url_id

class IWCT(nn.Module):
    
    def __init__(self,img_size:int = 512, folder_name='face_color_transfer', ckpt_name='G_innerface_128_512_145k.pt', force=False, device='cuda'):
        '''
        Related Links
        --------
        https://github.com/jmliu88/HeSer

        Methods
        --------
        - forward
            - Input
                - source_rgb
                    - dtype : tensor
                    - shape : (b, 3, 512, 512)
                    - min max : (-1, 1)
                - target_rgb  
                    - if you only have gray scale image, you should repeat gray scale image's channel to fit amount of rgb channel amount
                    - dtype : tensor
                    - shape : (b, 3, 512, 512)
                    - min max : (-1, 1)
                - source_onehot
                    - dtype : tensor
                    - shape : (b, 19, 512, 512)
                    - min max : (0 or 1)
                - target_onehot
                    - dtype : tensor
                    - shape : (b, 19, 512, 512)
                    - min max : (0 or 1)
                - target_gray
                    - dtype : tensor
                    - shape : (b, 1, 512, 512)
                    - min max : (-1, 1)
                - target_face_mask
                    - dtype : tensor
                    - shape : (b, 1, 512, 512)
                    - min max : (0 or 1)
                
            - Output
                - result
                    - shape : (1, 3, 512, 512)
                    - min max : (-1, 1)
                - color reference map
                    - shape : (1, 3, 512, 512))
                    - min max : (-1, 1)
        
        '''
        super(IWCT, self).__init__()
        self.img_size = img_size
        self.device = device
        self.generator = MyGenerator().to(self.device)
        
        url_id = get_url_id('~/.invz_pack/', folder_name, ckpt_name)
        root = os.path.join('~/.invz_pack/', folder_name)
        ckpt_path = check_ckpt_exist(root, ckpt_name = ckpt_name, url_id = url_id, force = force)
        ckpt = torch.load(ckpt_path, map_location=self.device)

        self.generator.load_state_dict(ckpt['model'])
        for param in self.generator.parameters():
            param.requires_grad = False
        self.generator.eval()
        del ckpt
        
    def forward(self, source_rgb, target_rgb, source_onehot, target_onehot, target_gray, target_face_mask, output_size=None):
        """
        Input
        ---------
            - source_rgb
                - dtype : tensor
                - shape : (b, 3, 512, 512)
                - min max : (-1, 1)
            - target_rgb  
                - if you only have gray scale image, you should repeat gray scale image's channel to fit amount of rgb channel amount
                - dtype : tensor
                - shape : (b, 3, 512, 512)
                - min max : (-1, 1)
            - source_onehot
                - dtype : tensor
                - shape : (b, 19, 512, 512)
                - min max : (0 or 1)
            - target_onehot
                - dtype : tensor
                - shape : (b, 19, 512, 512)
                - min max : (0 or 1)
            - target_gray
                - dtype : tensor
                - shape : (b, 1, 512, 512)
                - min max : (-1, 1)
            - target_face_mask
                - dtype : tensor
                - shape : (b, 1, 512, 512)
                - min max : (0 or 1)
            
        Output
        ---------
            - result
                - shape : (1, 3, 512, 512)
                - min max : (-1, 1)
            - color reference map
                - shape : (1, 3, 512, 512))
                - min max : (-1, 1)
        """
        _, _, origin_h, origin_w = source_rgb.shape
        source_rgb = F.interpolate(source_rgb, (self.img_size, self.img_size), mode='bilinear')
        target_rgb = F.interpolate(target_rgb, (self.img_size, self.img_size), mode='bilinear')
        source_onehot = F.interpolate(source_onehot, (self.img_size, self.img_size), mode='nearest')
        target_onehot = F.interpolate(target_onehot, (self.img_size, self.img_size), mode='nearest')
        target_gray = F.interpolate(target_gray, (self.img_size, self.img_size), mode='bilinear')
        target_face_mask = F.interpolate(target_face_mask, (self.img_size, self.img_size), mode='nearest')
        
        result, color_reference_map = self.generator(source_rgb, target_rgb, source_onehot, target_onehot, target_gray, target_face_mask)
        if output_size == None:
            result = F.interpolate(result, (origin_h, origin_w), mode='bilinear')
            color_reference_map = F.interpolate(color_reference_map, (origin_h, origin_w), mode='bilinear')
        else:
            result = F.interpolate(result, (output_size, output_size), mode='bilinear')
            color_reference_map = F.interpolate(color_reference_map, (output_size, output_size), mode='bilinear')
        
        return result, color_reference_map


import cv2
import numpy as np
from PIL import Image

import torch
import torch.nn.functional as F

from innerverz_package import Data_Process
from torchvision.transforms.functional import to_tensor, rgb_to_grayscale

DP = Data_Process()

image_path = 'sample.png'
pil_image = Image.open(image_path)
ts_image = DP.image_pp(pil_image, 512, normalize=True)


"""
        face alignment usage
"""
from innerverz_package import FaceAligner
FA = FaceAligner(align_style='ffhq')

cv2_image = cv2.imread(image_path)
aligned_face, tfm, tfm_inv, lms_5, lms_106, face_bool = FA.get_face(cv2_image)

face_bool, lms_106, lms_5 = FA.detect_lmk(cv2_image)

aligned_face, tfm, tfm_inv = FA.align_face_from_5(cv2_image, lms_5)

aligned_lmk = FA.affine_transform(lms_106, tfm)

    
"""
        FaceParser usage
"""
from innerverz_package import FaceParser
FP = FaceParser()

ts_aligned_face = DP.image_pp(aligned_face, 512, normalize=True)
label = FP.get_label(ts_aligned_face)
pp_label = DP.vis_pp(label, normalize=False, color=False)

"""
            Relighting usage
"""
from innerverz_package import ReLighter
RL = ReLighter()

from innerverz_package import IdExtractor
IE = IdExtractor()

from innerverz_package import DECA
deca = DECA()

ts_aligned_face = DP.image_pp(aligned_face, 512, normalize=True)
id_vector = IE(ts_aligned_face)

# pp
color_aligned_face = DP.image_pp(aligned_face, 512, False, True, False, 'cpu')
convexhull_label = DP.label_pp(pp_label, 'convexhull', aligned_lmk/2, 512, one_hot=False, batch=False, device='cpu')
convexhull_mask = np.where(convexhull_label.squeeze() != 0 ,1 , 0)
color_aligned_face = np.array(cv2.resize(aligned_face,(512,512)) * convexhull_mask[:,:,None])
convexhull_mask = torch.tensor(convexhull_mask).unsqueeze(0).unsqueeze(0).to('cuda')

# deca
source_image_dict = deca.data_preprocess(color_aligned_face, aligned_lmk/2)
source_tensor_images = source_image_dict['image'][None,...]
source_code_dict = deca.encode(source_tensor_images)

target_image_dict = deca.data_preprocess(color_aligned_face, aligned_lmk/2)
target_tensor_images = target_image_dict['image'][None,...]
target_code_dict = deca.encode(target_tensor_images)

target_code_dict['light'] = source_code_dict['light']

tform = target_image_dict['tform'][None, ...]
tform = torch.inverse(tform).transpose(1,2)
vis_dict = deca.decode(target_code_dict, original_image=target_image_dict['original_image'][None, ...] ,tform=tform)

target_masked_gray_face = rgb_to_grayscale((F.interpolate(target_image_dict['original_image'].unsqueeze(0).to('cuda'), (512, 512), mode='bilinear') - .5) * 2)
_target_masked_gray_face = target_masked_gray_face * convexhull_mask
relight_detail_shape = rgb_to_grayscale((F.interpolate(vis_dict['shape_detail_images'], (512,512), mode='bilinear') - .5) * 2, num_output_channels=1)
_relight_detail_shape = relight_detail_shape * convexhull_mask

relight_result, res_result = RL(_target_masked_gray_face, _relight_detail_shape, target_code_dict, id_vector)
_relight_result = DP.vis_pp(relight_result, True, True)
_res_result = DP.vis_pp(res_result, True, True)


import os
cwd = os.path.dirname(os.path.realpath(__file__))

import torch
import torch.nn as nn
import numpy as np
from .nets import MyGenerator
from ..utils import check_ckpt_exist, convert_image_type, get_url_id, get_grad_mask, get_center_coord

class FaceEnhancer(nn.Module):
    def __init__(self, folder_name='face_enhancer', ckpt_name = 'ckpt.zip', ckpt_face = 'face_090_395k.pt', ckpt_eye = 'eye_012_300k.pt', ckpt_mouth = 'mouth_007_80k.pt', force=False, device = 'cuda'):
        """
        Methods
        ---------
        - forward
            - Input
                - lmks
                    - dtype : numpy array
                    - shape : (b 106 2)
                - image
                    - dtype : tensor
                    - shape : (b 3 1024 1024)
                    - min max : (-1 1)
            - Output
                - dtype : tensor
                - shape : (b 3 1024 1024)
                - min max : (-1 1)
        """
        super(FaceEnhancer, self).__init__()

        self.device = device
        url_id = get_url_id('~/.invz_pack/', folder_name, ckpt_name)
        root = os.path.join('~/.invz_pack/', folder_name)
        self.dir_folder_path = check_ckpt_exist(root, ckpt_name = ckpt_name, url_id = url_id, force = force)[:-4]
        
        self.face_enhancer = MyGenerator().to(device)
        self.eye_enhancer = MyGenerator().to(device)
        self.mouth_enhancer = MyGenerator().to(device)
        
        ckpt_pairs = [
            [self.face_enhancer, os.path.join(self.dir_folder_path, ckpt_face)],
            [self.eye_enhancer, os.path.join(self.dir_folder_path, ckpt_eye)],
            [self.mouth_enhancer, os.path.join(self.dir_folder_path, ckpt_mouth)],
        ]
        for enhancer, ckpt_path in ckpt_pairs:
            ckpt = torch.load(os.path.join(cwd, ckpt_path), map_location=device)
            enhancer.load_state_dict(ckpt['model'], strict=False)
            for param in enhancer.parameters():
                param.requires_grad = False
            enhancer.eval()
            del ckpt

        self.grad_mask = torch.from_numpy(get_grad_mask()).to(device)

    def forward(self, lmks, imgs):
        batch_num = imgs.size()[0]

        full_result = self.face_enhancer(imgs) # size: 1024, value range: [-1, 1]

        target_R_eye = torch.zeros((batch_num, 3, 256, 256), device=self.device)
        target_L_eye = torch.zeros((batch_num, 3, 256, 256), device=self.device) 
        target_Mouth = torch.zeros((batch_num, 3, 256, 256), device=self.device) 

        for idx in range(batch_num):
            L_xc, L_yc, _, _ = get_center_coord(lmks[idx], 'L_eye')
            R_xc, R_yc, _, _ = get_center_coord(lmks[idx], 'R_eye')
            M_xc, M_yc, _, _ = get_center_coord(lmks[idx], 'mouth')

            target_L_eye[idx] = full_result[idx, :, L_yc-128:L_yc+128, L_xc-128:L_xc+128]
            target_R_eye[idx] = full_result[idx, :, R_yc-128:R_yc+128, R_xc-128:R_xc+128]
            target_Mouth[idx] = full_result[idx, :, M_yc-128:M_yc+128, M_xc-128:M_xc+128]

        L_eye_result = self.eye_enhancer(target_L_eye)
        R_eye_result = self.eye_enhancer(target_R_eye)
        Mouth_result = self.mouth_enhancer(target_Mouth)

        full_result[:, :, L_yc-128:L_yc+128, L_xc-128:L_xc+128] = L_eye_result*self.grad_mask + full_result[:, :, L_yc-128:L_yc+128, L_xc-128:L_xc+128]*(1-self.grad_mask)
        full_result[:, :, R_yc-128:R_yc+128, R_xc-128:R_xc+128] = R_eye_result*self.grad_mask + full_result[:, :, R_yc-128:R_yc+128, R_xc-128:R_xc+128]*(1-self.grad_mask)
        full_result[:, :, M_yc-128:M_yc+128, M_xc-128:M_xc+128] = Mouth_result*self.grad_mask + full_result[:, :, M_yc-128:M_yc+128, M_xc-128:M_xc+128]*(1-self.grad_mask)

        return full_result



import os
import sys

import torch
import torch.nn as nn

from .net import MyGenerator
from ..utils import check_ckpt_exist, get_url_id

class Reage(nn.Module):
    def __init__(self, folder_name='reage', ckpt_name = 'G_512_130k.pt', force=False, device = 'cuda'):
        """
        Related Links
        --------
        https://studios.disneyresearch.com/2022/11/30/production-ready-face-re-aging-for-visual-effects/
        
        Options
        --------
        ckpt_name
            - G_512_130k.pt : image size(512)
            - G_1024_110k.pt : image size(1024)
        
        Forwards
        --------
        - face image
            - aligned face with ffhq rules
            - dtype : tensor
            - shape : (b, 3, h, w)
            - min max : (-1 1)
        
        - mask 
            - skin(1) + brow(2,3) + ears(7,8) + nose(10) + mouth(11) + lip(12,13) + neck(14) area
            - dtype : tensor
            - shape : (b, 1, h, w)
            - min max : (0 or 1)
        """
        super(Reage, self).__init__()
        self.device = device
        self.generator = MyGenerator().to(self.device)
        
        url_id = get_url_id('~/.invz_pack/', folder_name, ckpt_name)
        root = os.path.join('~/.invz_pack/', folder_name)
        ckpt_path = check_ckpt_exist(root, ckpt_name = ckpt_name, url_id = url_id, force = force)
        ckpt = torch.load(ckpt_path, map_location=self.device)

        self.generator.load_state_dict(ckpt['model'])
        for param in self.generator.parameters():
            param.requires_grad = False
        self.generator.eval()
        del ckpt
        
    def forward(self, masked_image, mask, from_age, to_age):
        
        model_input = torch.cat((masked_image, mask * from_age / 100, mask * to_age / 100), dim=1)
        
        aging_delta = self.generator(model_input)
        result = masked_image + aging_delta
        
        return result, aging_delta


'''
crop
for torch tensor
Given image, bbox(center, bboxsize)
return: cropped image, tform(used for transform the keypoint accordingly)
only support crop to squared images
'''
import torch
from kornia.geometry.transform.imgwarp import (
    warp_perspective, get_perspective_transform, warp_affine
)

def points2bbox(points, points_scale=None):
    if points_scale:
        assert points_scale[0]==points_scale[1]
        points = points.clone()
        points[:,:,:2] = (points[:,:,:2]*0.5 + 0.5)*points_scale[0]
    min_coords, _ = torch.min(points, dim=1)
    xmin, ymin = min_coords[:, 0], min_coords[:, 1]
    max_coords, _ = torch.max(points, dim=1)
    xmax, ymax = max_coords[:, 0], max_coords[:, 1]
    center = torch.stack([xmax + xmin, ymax + ymin], dim=-1) * 0.5

    width = (xmax - xmin)
    height = (ymax - ymin)
    # Convert the bounding box to a square box
    size = torch.max(width, height).unsqueeze(-1)
    return center, size

def augment_bbox(center, bbox_size, scale=[1.0, 1.0], trans_scale=0.):
    batch_size = center.shape[0]
    trans_scale = (torch.rand([batch_size, 2], device=center.device)*2. -1.) * trans_scale
    center = center + trans_scale*bbox_size # 0.5
    scale = torch.rand([batch_size,1], device=center.device) * (scale[1] - scale[0]) + scale[0]
    size = bbox_size*scale
    return center, size

def crop_tensor(image, center, bbox_size, crop_size, interpolation = 'bilinear', align_corners=False):
    ''' for batch image
    Args:
        image (torch.Tensor): the reference tensor of shape BXHxWXC.
        center: [bz, 2]
        bboxsize: [bz, 1]
        crop_size;
        interpolation (str): Interpolation flag. Default: 'bilinear'.
        align_corners (bool): mode for grid_generation. Default: False. See
          https://pytorch.org/docs/stable/nn.functional.html#torch.nn.functional.interpolate for details
    Returns:
        cropped_image
        tform
    '''
    dtype = image.dtype
    device = image.device
    batch_size = image.shape[0]
    # points: top-left, top-right, bottom-right, bottom-left    
    src_pts = torch.zeros([4,2], dtype=dtype, device=device).unsqueeze(0).expand(batch_size, -1, -1).contiguous()

    src_pts[:, 0, :] = center - bbox_size*0.5  # / (self.crop_size - 1)
    src_pts[:, 1, 0] = center[:, 0] + bbox_size[:, 0] * 0.5
    src_pts[:, 1, 1] = center[:, 1] - bbox_size[:, 0] * 0.5
    src_pts[:, 2, :] = center + bbox_size * 0.5
    src_pts[:, 3, 0] = center[:, 0] - bbox_size[:, 0] * 0.5
    src_pts[:, 3, 1] = center[:, 1] + bbox_size[:, 0] * 0.5

    DST_PTS = torch.tensor([[
        [0, 0],
        [crop_size - 1, 0],
        [crop_size - 1, crop_size - 1],
        [0, crop_size - 1],
    ]], dtype=dtype, device=device).expand(batch_size, -1, -1)
    # estimate transformation between points
    dst_trans_src = get_perspective_transform(src_pts, DST_PTS)
    # simulate broadcasting
    # dst_trans_src = dst_trans_src.expand(batch_size, -1, -1)

    # warp images 
    cropped_image = warp_affine(
        image, dst_trans_src[:, :2, :], (crop_size, crop_size),
        flags=interpolation, align_corners=align_corners)

    tform = torch.transpose(dst_trans_src, 2, 1)
    # tform = torch.inverse(dst_trans_src)
    return cropped_image, tform

class Cropper(object):
    def __init__(self, crop_size, scale=[1,1], trans_scale = 0.):
        self.crop_size = crop_size
        self.scale = scale
        self.trans_scale = trans_scale

    def crop(self, image, points, points_scale=None):
        # points to bbox
        center, bbox_size = points2bbox(points.clone(), points_scale)
        # argument bbox. TODO: add rotation?
        center, bbox_size = augment_bbox(center, bbox_size, scale=self.scale, trans_scale=self.trans_scale)
        # crop
        cropped_image, tform = crop_tensor(image, center, bbox_size, self.crop_size)
        return cropped_image, tform
    
    def transform_points(self, points, tform, points_scale=None, normalize = True):
        points_2d = points[:,:,:2]
        
        #'input points must use original range'
        if points_scale:
            assert points_scale[0]==points_scale[1]
            points_2d = (points_2d*0.5 + 0.5)*points_scale[0]

        batch_size, n_points, _ = points.shape
        trans_points_2d = torch.bmm(
                        torch.cat([points_2d, torch.ones([batch_size, n_points, 1], device=points.device, dtype=points.dtype)], dim=-1), 
                        tform
                        ) 
        trans_points = torch.cat([trans_points_2d[:,:,:2], points[:,:,2:]], dim=-1)
        if normalize:
            trans_points[:,:,:2] = trans_points[:,:,:2]/self.crop_size*2 - 1
        return trans_points

def transform_points(points, tform, points_scale=None, out_scale=None):
    points_2d = points[:,:,:2]
        
    #'input points must use original range'
    if points_scale:
        assert points_scale[0]==points_scale[1]
        points_2d = (points_2d*0.5 + 0.5)*points_scale[0]
    # import ipdb; ipdb.set_trace()

    batch_size, n_points, _ = points.shape
    trans_points_2d = torch.bmm(
                    torch.cat([points_2d, torch.ones([batch_size, n_points, 1], device=points.device, dtype=points.dtype)], dim=-1), 
                    tform
                    ) 
    if out_scale: # h,w of output image size
        trans_points_2d[:,:,0] = trans_points_2d[:,:,0]/out_scale[1]*2 - 1
        trans_points_2d[:,:,1] = trans_points_2d[:,:,1]/out_scale[0]*2 - 1
    trans_points = torch.cat([trans_points_2d[:,:,:2], points[:,:,2:]], dim=-1)
    return trans_points


from innerverz import Reage
reage = Reage()

from innerverz import Data_Process
DP = Data_Process()

from innerverz import FaceParser
FP = FaceParser()

from innerverz import FaceAligner
FA = FaceAligner(512, 'ffhq')

import cv2
import numpy as np
from PIL import Image

import torch

image = cv2.imread('./assets/assets/000.png')
face,_,_,_,_,_ = FA.get_face(image)
ts_face = DP.image_pp(face, 512, normalize=True)
ts_label = FP.get_label(ts_face)
ts_label = ts_label.unsqueeze(0)
ts_mask = torch.where(ts_label<15, 1, 0)
ts_mask -= torch.where(ts_label==0, 1, 0)
ts_mask -= torch.where(ts_label==4, 1, 0)
ts_mask -= torch.where(ts_label==5, 1, 0)

result,_ = reage(ts_face, ts_mask, 40, 20)
vis_result = DP.vis_pp(result)
cv2.imwrite('./result.png', vis_result)


import torch
import torch.nn as nn
import torch.nn.functional as F

class AdaIN(nn.Module):
    def __init__(self, style_dim, num_features):
        super().__init__()
        self.norm = nn.InstanceNorm2d(num_features, affine=False)
        self.fc = nn.Linear(style_dim, num_features*2)

    def forward(self, x, s):
        h = self.fc(s)
        h = h.view(h.size(0), h.size(1), 1, 1)
        gamma, beta = torch.chunk(h, chunks=2, dim=1)
        return (1 + gamma) * self.norm(x) + beta

class AdaINResBlock(nn.Module):
    def __init__(self, in_c, out_c, scale_factor=1, style_dim=512):
        super(AdaINResBlock, self).__init__()

        self.conv1 = nn.Conv2d(in_channels=in_c, out_channels=out_c, kernel_size=3, stride=1, padding=1, padding_mode='reflect', bias=False)
        self.AdaIN1 = AdaIN(style_dim, out_c)
        self.activ1 = nn.LeakyReLU(0.2)

        self.conv2 = nn.Conv2d(in_channels=out_c, out_channels=out_c, kernel_size=3, stride=1, padding=1, padding_mode='reflect', bias=False)
        self.AdaIN2 = AdaIN(style_dim, out_c)
        self.activ2 = nn.LeakyReLU(0.2)

        self.conv1x1 = nn.Conv2d(in_channels=in_c, out_channels=out_c, kernel_size=1, stride=1, padding=0, padding_mode='reflect', bias=False)

        self.scale_factor = scale_factor
        self.resize = scale_factor != 1

    def forward(self, feat, style):

        feat1 = feat
        feat1 = self.conv1(feat1)
        feat1 = self.AdaIN1(feat1, style)
        feat1 = self.activ1(feat1)

        if self.resize:
            feat1 = F.interpolate(feat1, scale_factor=self.scale_factor, mode='bilinear')

        feat1 = self.conv2(feat1)
        feat1 = self.AdaIN2(feat1, style)
        feat1 = self.activ2(feat1)

        # skip connction
        feat2 = feat
        if self.resize:
            feat2 = F.interpolate(feat2, scale_factor=self.scale_factor, mode='bilinear') # size
            feat2 = self.conv1x1(feat2) # chnnel dim

        return feat1 + feat2


class MLP(nn.Module):
    # TODO: input_nc check!
    def __init__(self, input_nc=3, output_nc=512):
        super(MLP, self).__init__()
        
        self.linear_0 = nn.Linear(input_nc, output_nc)
        self.linear_1 = nn.Linear(output_nc, output_nc)
        self.linear_2 = nn.Linear(output_nc, output_nc)
        self.linear_3 = nn.Linear(output_nc, output_nc)
        
    def forward(self, input):
        output = self.linear_0(input)
        output = self.linear_1(output)
        output = self.linear_2(output)
        output = self.linear_3(output)
        
        return output

class Unet(nn.Module):
    def __init__(self, input_nc=3, output_nc=3):
        super(Unet, self).__init__()
        
        self.input_layer = nn.Sequential(nn.ReflectionPad2d(3), nn.Conv2d(input_nc, 32, kernel_size=7))

        self.down1 = AdaINResBlock(32, 64, scale_factor=.5)
        self.down2 = AdaINResBlock(64, 128, scale_factor=.5)
        self.down3 = AdaINResBlock(128, 256, scale_factor=.5)
        self.down4 = AdaINResBlock(256, 512, scale_factor=.5)
        
        self.bottle_neck = AdaINResBlock(512, 512, scale_factor=1)
                           
        self.up4 = AdaINResBlock(1024, 256, scale_factor=2)
        self.up3 = AdaINResBlock(512, 128, scale_factor=2)
        self.up2 = AdaINResBlock(256, 64, scale_factor=2)
        self.up1 = AdaINResBlock(128, 32, scale_factor=2)

        self.output_layer = nn.Sequential(nn.ReflectionPad2d(3), nn.Conv2d(32, output_nc, kernel_size=7))

    def forward(self, x, style):

        x_in = self.input_layer(x)

        x_d1 = self.down1(x_in, style)
        x_d2 = self.down2(x_d1, style)
        x_d3 = self.down3(x_d2, style)
        x_d4 = self.down4(x_d3, style)

        x_bn = self.bottle_neck(x_d4, style)

        x_u4 = self.up4(torch.cat([x_d4,x_bn], dim=1), style)
        x_u3 = self.up3(torch.cat([x_d3,x_u4], dim=1), style)
        x_u2 = self.up2(torch.cat([x_d2,x_u3], dim=1), style)
        x_u1 = self.up1(torch.cat([x_d1,x_u2], dim=1), style)

        x_out = self.output_layer(x_u1)

        return x_out
        # return F.tanh(x_out) * 2

class MyGenerator(nn.Module):
    def __init__(self):
        super(MyGenerator, self).__init__()
        self.mlp = MLP(36+128+512,512)
        self.unet = Unet(2, 1)

    def forward(self, target_face, source_shape, source_params):
        input_images = torch.cat([target_face, source_shape], dim=1)
        
        input_params = self.mlp(source_params)
        outputs = self.unet(input_images, input_params)

        return outputs
    
